@@ -1,8 +1,7 @@
-#! /bin/sh
-# -*- tcl -*- \
-exec tclsh8.5 "$0" ${1+"$@"}
+#! /usr/bin/env tclsh
+## -*- tcl -*-
 
-package require Tcl 8.4
+package require Tcl 8.4-
 package require Tk
 
 package require Plotchart
@@ -15,10 +14,10 @@ pack [canvas .c] -fill both -side top
 
 set p [::Plotchart::createBoxplot .c {0 40 5} {A B C D E F}]
 
-$p plot A {0 1 2 5 7 1 4 5 0.6 5 5.5}
-$p plot C {2 2 3 6 1.5 3}
+$p plot data A {0 1 2 5 7 1 4 5 0.6 5 5.5}
+$p plot data C {2 2 3 6 1.5 3}
 
-$p plot E {2 3 3 4 7 8 9 9 10 10 11 11 11 14 15 17 17 20 24 29}
+$p plot data E {2 3 3 4 7 8 9 9 10 10 11 11 11 14 15 17 17 20 24 29}
 
 #
 # Demonstration of selected x labels - for version 1.6.2
